package Search::Query::Dialect::DBIxClass::Fuzzy;
# ABSTRACT: Dialect::DBIxClass and data_type numeric and time supports
$Search::Query::Dialect::DBIxClass::Fuzzy::VERSION = '0.201410270';
use Moo;
extends 'Search::Query::Dialect::DBIxClass';

use DateTime::Format::Natural;
use List::Flatten;
use List::Util qw(reduce);

has result_source => ( is => 'ro', required => 1 );

sub _dbic_op_date      { shift->_dbic_op_datetime(@_) }
sub _dbic_op_timestamp { shift->_dbic_op_datetime(@_) }

sub _dbic_op_integer   { shift->_dbic_op_numeric(@_) }
sub _dbic_op_smallint  { shift->_dbic_op_numeric(@_) }
sub _dbic_op_boolean   { shift->_dbic_op_numeric(@_) }

sub _dbic_op_text      { shift->_dbic_op_fallback(@_) }
sub _dbic_op_varchar   { shift->_dbic_op_fallback(@_) }

*_wrap_dbic_ops = \&Search::Query::Dialect::DBIxClass::_wrap_dbic_ops;

my %negated_prefix = (
    '+' => '-',
    '-' => '+',
);

my %negated_op = (
    ':'  => '!~',
    '~'  => '!~',
    '='  => '!=',
    '==' => '!=',
    '>'  => '<=',
    '<'  => '>=',
    '#'  => '!#',
    '()' => '()',
);

# add the negated negations
# we don't care if a hash key gets overwritten because they are just different
# chars for the same op
%negated_op = ( %negated_op, reverse %negated_op );

sub _dbic_op {
    my ( $self, $clause, $prefix, $colnames ) = @_;

    _wrap_dbic_ops(
        flat map {
            my $colname = $_;

            my @rels;
            my $result_source = $self->result_source;
            while ($colname =~ s/^(.*?)\.//) {
                push @rels, $1;
                $result_source = $result_source->related_source($1);
            }

            my $code;
            if ($result_source->has_column($colname)) {
                my $info = $result_source->column_info($colname);
                $code = $self->can("_dbic_op_$info->{data_type}");
            }
            $code ||= $self->can('_dbic_op_fallback');

            my $join = reduce { +{ $a => $b } } @rels;
            if ($join) {
                $self->{join} = _merge_join($self->{join}, $join);
            }

            $self->$code($clause, $prefix, [ "@{[ $rels[-1]||'me' ]}.$colname" ]);
        } @$colnames
    );
}

sub _datetime {
    my ($self, $str) = @_;

    my $parser = DateTime::Format::Natural->new;
    my $dtf = $self->result_source->schema->storage->datetime_parser;
    my $dt = $parser->parse_datetime($str);
    if ($parser->success) {
        $dtf->format_datetime($dt);
    }
    else {
        die "Invalid datetime $str";
    }
}

sub _dbic_op_fallback {
    my ( $self, $clause, $prefix, $colnames ) = @_;
    die 'no clause'
        unless defined $clause;
    die 'no prefix'
        unless defined $prefix;
    die 'column names required'
        unless defined $colnames;

    # normalize operator
    my $op = $clause->{op} || ":";

    # the - prefix inverts the ops
    if ( $prefix eq '-' ) {
        $op =
            exists $negated_op{$op}
            ? $negated_op{$op}
            : die "negated op for '$op' not handled by DBIxClass dialect";
    }

    if ( $op eq '()' ) {
        return $self->as_dbic_query( $clause->{value}, $prefix );
    }
    elsif ( $op eq ':' || $op eq '~' ) {
        return _wrap_dbic_ops(
            map {
                # TEXT()はenum向け
                \[  "LOWER(TEXT($_)) LIKE ?",
                    [ plain_value => lc "%$clause->{value}%" ]
                ]
            } @$colnames
        );
    }
    elsif ( $op eq '!~' ) {
        return _wrap_dbic_ops(
            map {
                # TEXT()はenum向け
                \[  "COALESCE( LOWER(TEXT($_)), '' ) NOT LIKE ?",
                    [ plain_value => lc "%$clause->{value}%" ]
                ]
            } @$colnames
        );
    }
    elsif ( $op eq '=' || $op eq '==' ) {
        return _wrap_dbic_ops(
            map {
                { $_ => $clause->{value} }
            } @$colnames
        );
    }
    elsif ( $op eq '>' || $op eq '<' || $op eq '>=' || $op eq '<=' ) {
        return _wrap_dbic_ops(
            map {
                { $_ => { $op => $clause->{value} } }
            } @$colnames
        );
    }
    elsif ( $op eq '#' ) {
        return _wrap_dbic_ops(
            map {
                { $_ => { -in => [ split( /,/, $clause->{value} ) ] } }
            } @$colnames
        );
    }
    elsif ( $op eq '!#' ) {
        return _wrap_dbic_ops(
            map {
                { $_ => { -not_in => [ split( /,/, $clause->{value} ) ] } }
            } @$colnames
        );
    }
    else {
        die "operator '$op' not supported by DBIxClass dialect";
    }
}

sub _dbic_op_datetime {
    my ( $self, $clause, $prefix, $colnames ) = @_;
    die 'no clause'
        unless defined $clause;
    die 'no prefix'
        unless defined $prefix;
    die 'column names required'
        unless defined $colnames;

    # normalize operator
    my $op = $clause->{op} || ":";

    # the - prefix inverts the ops
    if ( $prefix eq '-' ) {
        $op =
            exists $negated_op{$op}
            ? $negated_op{$op}
            : die "negated op for '$op' not handled by DBIxClass dialect";
    }

    if ( $op eq '()' ) {
        return $self->as_dbic_query( $clause->{value}, $prefix );
    }
    elsif ( $op eq ':' || $op eq '~' ) {
        return _wrap_dbic_ops(
            map {
                \[  "TEXT($_) LIKE ?",
                    [ plain_value => "%$clause->{value}%" ]
                    ]
            } @$colnames
        );
    }
    elsif ( $op eq '!~' ) {
        return _wrap_dbic_ops(
            map {
                \[  "COALESCE( TEXT($_), '' ) NOT LIKE ?",
                    [ plain_value => "%$clause->{value}%" ]
                    ]
            } @$colnames
        );
    }
    elsif ( $op eq '=' || $op eq '==' ) {
        return _wrap_dbic_ops(
            map {
                { $_ => $self->_datetime($clause->{value}) }
            } @$colnames
        );
    }
    elsif ( $op eq '>' || $op eq '<' || $op eq '>=' || $op eq '<=' ) {
        return _wrap_dbic_ops(
            map {
                { $_ => { $op => $self->_datetime($clause->{value}) } }
            } @$colnames
        );
    }
    elsif ( $op eq '#' ) {
        return _wrap_dbic_ops(
            map {
                { $_ => { -in => [ split( /,/, $self->_datetime($clause->{value}) ) ] } }
            } @$colnames
        );
    }
    elsif ( $op eq '!#' ) {
        return _wrap_dbic_ops(
            map {
                { $_ => { -not_in => [ split( /,/, $self->_datetime($clause->{value}) ) ] } }
            } @$colnames
        );
    }
    else {
        die "operator '$op' not supported by DBIxClass dialect";
    }
}

sub _dbic_op_numeric {
    my ( $self, $clause, $prefix, $colnames ) = @_;
    die 'no clause'
        unless defined $clause;
    die 'no prefix'
        unless defined $prefix;
    die 'column names required'
        unless defined $colnames;

    # normalize operator
    my $op = $clause->{op} || ":";

    # the - prefix inverts the ops
    if ( $prefix eq '-' ) {
        $op =
            exists $negated_op{$op}
            ? $negated_op{$op}
            : die "negated op for '$op' not handled by DBIxClass dialect";
    }

    if ( $op eq '()' ) {
        return $self->as_dbic_query( $clause->{value}, $prefix );
    }
    elsif ( $op eq ':' || $op eq '~' ) {
        return _wrap_dbic_ops(
            map {
                \[  "TEXT($_) LIKE ?",
                    [ plain_value => "%$clause->{value}%" ]
                    ]
            } @$colnames
        );
    }
    elsif ( $op eq '!~' ) {
        return _wrap_dbic_ops(
            map {
                \[  "COALESCE( TEXT($_), '' ) NOT LIKE ?",
                    [ plain_value => "%$clause->{value}%" ]
                    ]
            } @$colnames
        );
    }
    elsif ( $op eq '=' || $op eq '==' ) {
        return _wrap_dbic_ops(
            map {
                { $_ => $clause->{value} }
            } @$colnames
        );
    }
    elsif ( $op eq '>' || $op eq '<' || $op eq '>=' || $op eq '<=' ) {
        return _wrap_dbic_ops(
            map {
                { $_ => { $op => $clause->{value} } }
            } @$colnames
        );
    }
    elsif ( $op eq '#' ) {
        return _wrap_dbic_ops(
            map {
                { $_ => { -in => [ split( /,/, $clause->{value} ) ] } }
            } @$colnames
        );
    }
    elsif ( $op eq '!#' ) {
        return _wrap_dbic_ops(
            map {
                { $_ => { -not_in => [ split( /,/, $clause->{value} ) ] } }
            } @$colnames
        );
    }
    else {
        die "operator '$op' not supported by DBIxClass dialect";
    }
}

# https://github.com/abraxxa/Search-Query-Dialect-DBIxClass/pull/2
sub as_dbic_query {
    my $self         = shift;
    my $tree         = shift || $self;    # if called recursively by the () op
    my $query_prefix = shift || '+';

    # ensure default_field is always an arrayref
    my @default_fields =
        ( exists $self->parser->{default_field}
            && !ref $self->parser->{default_field} )
        ? [ $self->parser->{default_field} ]
        : @{ $self->parser->{default_field} };

    my @q;
    my $mandatory = 0;
    foreach my $prefix ( '+', '', '-' ) {
        next unless exists $tree->{$prefix};

        for my $clause ( @{ $tree->{$prefix} } ) {
            my @colnames =
                defined $clause->{field}
                ? ( $clause->{field} )
                : @default_fields;

            my $clause_prefix =
                  $query_prefix eq '-'
                ? $negated_prefix{$prefix}
                : $prefix;

            my $dbic_op =
                $self->_dbic_op( $clause, $clause_prefix, \@colnames );

            push @q, $dbic_op;
            $mandatory = 1 if $prefix eq '+' or $prefix eq '-';
        }
    }

    my %search_params    = ( ($mandatory ? '-and' : '-or') => \@q );
    my %search_attribute = ( join => $self->{join} );

    return \%search_params, \%search_attribute;
}

use feature ":5.10";
use Hash::Merge;
use List::MoreUtils qw(uniq);
use Params::Util qw(_HASH);
sub _merge_join {
    state $merge = do {
        my $merge = Hash::Merge->new;
        $merge->specify_behavior(
            {
                SCALAR => {
                    SCALAR => sub {
                        my @vals = uniq grep $_, $_[0], $_[1];
                        @vals > 1 ? \@vals : $vals[0];
                    },
                    ARRAY  => sub {
                        my ($scalar, $array) = @_;
                        my %keys = map { $_ => 1 } map { keys %$_ } grep { _HASH $_ } @$array;
                        my @vals = uniq( ( $keys{$scalar} ? () : $scalar ), @$array );
                        @vals > 1 ? \@vals : $vals[0];
                    },
                    HASH   => sub {
                        my ($scalar, $hash) = @_;
                        my %keys = map { $_ => 1 } keys %$hash;
                        my @vals = ( ( $scalar ? ($keys{$scalar} ? () : $scalar) : () ), $hash );
                        @vals > 1 ? \@vals : $vals[0];
                    },
                },
                ARRAY => {
                    SCALAR => sub { 
                        my ($array, $scalar) = @_;
                        my %keys = map { $_ => 1 } map { keys %$_ } grep { _HASH $_ } @$array;
                        my @vals = uniq( ( $keys{$scalar} ? () : $scalar ), @$array );
                        @vals > 1 ? \@vals : $vals[0];
                    },
                    ARRAY  => sub { [ @{$_[0]}, @{$_[1]} ] },
                    HASH   => sub { 
                        my ($array, $hash) = @_;
                        my @s;
                        my @h;
                        for (@$array) {
                            if (_HASH $_) {
                                push @h, $_;
                            }
                            else {
                                push @s, $_;
                            }
                        }
                        $hash = Hash::Merge::_merge_hashes($hash, $_) for @h;
                        [ (map { $hash->{$_} ? () : $_ } @s), $hash ];
                    },
                },
                HASH => {
                    SCALAR => sub {
                        my ($hash, $scalar) = @_;
                        my %keys = map { $_ => 1 } keys %$hash;
                        my @vals = ( ( $scalar ? ($keys{$scalar} ? () : $scalar) : () ), $hash );
                        @vals > 1 ? \@vals : $vals[0];
                    },
                    ARRAY  => sub {
                        my ($hash, $array) = @_;
                        my @s;
                        my @h;
                        for (@$array) {
                            if (_HASH $_) {
                                push @h, $_;
                            }
                            else {
                                push @s, $_;
                            }
                        }
                        $hash = Hash::Merge::_merge_hashes($hash, $_) for @h;
                        [ (map { $hash->{$_} ? () : $_ } @s), $hash ];
                    },
                    HASH   => sub { Hash::Merge::_merge_hashes( $_[0], $_[1] ) }, 
                },
            }, 
        );
        $merge;
    };

    if (defined $_[0]) {
        $merge->merge(@_);
    }
    else {
        $_[1];
    }
}

1;

__END__

=pod

=head1 NAME

Search::Query::Dialect::DBIxClass::Fuzzy - Dialect::DBIxClass and data_type numeric and time supports

=head1 VERSION

version 0.201410270

=head1 AUTHOR

Tomohiro Hosaka <bokutin@bokut.in>

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2014 by Tomohiro Hosaka.

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut
